import numpy as np
import tensorflow as tf


# Code from https://www.tensorflow.org/tutorials/load_data/tfrecord

def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()  # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def serialize_value(value):
    if type(value) == np.ndarray:
        return _bytes_feature(tf.io.serialize_tensor(value))
    elif type(value) == int:
        return _int64_feature(value)
    elif type(value) == float:
        return _float_feature(value)
    else:
        raise TypeError(f"Didn't recognize type {type(value)}")



import tensorflow as tf
from tensorflow.keras import layers as tf_layers

import nfp


class GraphLayer(tf_layers.Layer):
    """ Base class for all GNN layers """

    def __init__(self, dropout: float = 0.0, **kwargs):
        super().__init__(**kwargs)
        self.dropout = dropout
        self.supports_masking = True

    def build(self, input_shape):
        if len(input_shape) == 4:
            self.use_global = True
            self.tile = nfp.Tile()

        elif len(input_shape) == 3:
            self.use_global = False

        else:
            raise RuntimeError("wrong input shape")

        if self.dropout > 0.:
            self.dropout_layer = tf_layers.Dropout(self.dropout)

    def get_config(self):
        return {"dropout": self.dropout}


class EdgeUpdate(GraphLayer):
    def build(self, input_shape):
        """ inputs = [atom_state, bond_state, connectivity]
        shape(bond_state) = [batch, num_bonds, bond_features]
        """
        super().build(input_shape)

        self.gather = nfp.Gather()
        self.concat = nfp.ConcatDense()

    def call(self, inputs, mask=None, **kwargs):
        """ Inputs: [atom_state, bond_state, connectivity]
            Outputs: bond_state

        """
        if not self.use_global:
            atom_state, bond_state, connectivity = inputs
        else:
            atom_state, bond_state, connectivity, global_state = inputs
            global_state = self.tile([global_state, bond_state])

        # Get nodes at start and end of edge
        source_atom = self.gather([atom_state, connectivity[:, :, 0]])
        target_atom = self.gather([atom_state, connectivity[:, :, 1]])

        if not self.use_global:
            new_bond_state = self.concat(
                [bond_state, source_atom, target_atom])
        else:
            new_bond_state = self.concat(
                [bond_state, source_atom, target_atom, global_state])

        if self.dropout > 0.:
            new_bond_state = self.dropout_layer(new_bond_state)

        return new_bond_state

    def compute_output_shape(self, input_shape):
        return input_shape[1]

    def compute_mask(self, inputs, mask=None):
        if mask:
            return mask[1]
        else:
            return None


class NodeUpdate(GraphLayer):
    def build(self, input_shape):
        super().build(input_shape)

        num_features = input_shape[1][-1]

        self.gather = nfp.Gather()

        self.concat = nfp.ConcatDense()
        self.reduce = nfp.Reduce(reduction='sum')

        self.dense1 = tf_layers.Dense(2 * num_features, activation='relu')
        self.dense2 = tf_layers.Dense(num_features)

    def call(self, inputs, mask=None, **kwargs):
        """ Inputs: [atom_state, bond_state, connectivity]
            Outputs: atom_state

        """
        if not self.use_global:
            atom_state, bond_state, connectivity = inputs
        else:
            atom_state, bond_state, connectivity, global_state = inputs
            global_state = self.tile([global_state, bond_state])

        source_atom = self.gather([atom_state, connectivity[:, :, 1]])

        if not self.use_global:
            messages = self.concat([source_atom, bond_state])
        else:
            messages = self.concat([source_atom, bond_state, global_state])

        if mask is not None:
            # Only works for sum, max
            messages = tf.where(tf.expand_dims(mask[1], axis=-1), messages,
                                tf.zeros_like(messages))

        new_atom_state = self.reduce(
            [messages, connectivity[:, :, 0], atom_state])

        # Dense net after message reduction
        new_atom_state = self.dense1(new_atom_state)
        new_atom_state = self.dense2(new_atom_state)

        if self.dropout > 0.:
            new_atom_state = self.dropout_layer(new_atom_state)

        return new_atom_state

    def compute_output_shape(self, input_shape):
        return input_shape[0]

    def compute_mask(self, inputs, mask=None):
        if mask:
            return mask[0]
        else:
            return None


class GlobalUpdate(GraphLayer):
    def __init__(self, units, num_heads, **kwargs):
        super().__init__(**kwargs)
        self.units = units  # H
        self.num_heads = num_heads  # N
        self.supports_masking = False

    def build(self, input_shape):
        super().build(input_shape)
        dense_units = self.units * self.num_heads  # N*H
        self.query_layer = tf_layers.Dense(self.num_heads, name='query')
        self.value_layer = tf_layers.Dense(dense_units, name='value')

    def transpose_scores(self, input_tensor):
        input_shape = tf.shape(input_tensor)
        output_shape = [
            input_shape[0], input_shape[1], self.num_heads, self.units
        ]
        output_tensor = tf.reshape(input_tensor, output_shape)
        return tf.transpose(a=output_tensor, perm=[0, 2, 1, 3])  # [B,N,S,H]

    def call(self, inputs, mask=None, **kwargs):
        if not self.use_global:
            atom_state, bond_state, connectivity = inputs
        else:
            atom_state, bond_state, connectivity, global_state = inputs

        batch_size = tf.shape(atom_state)[0]

        graph_elements = tf.concat([atom_state, bond_state], axis=1)
        query = self.query_layer(graph_elements)  # [B,N,S,H]

        if mask is not None:
            graph_element_mask = tf.concat([mask[0], mask[1]], axis=1)
            query = tf.where(tf.expand_dims(graph_element_mask, axis=-1),
                             query,
                             tf.ones_like(query) * query.dtype.min)

        query = tf.transpose(query, perm=[0, 2, 1])
        value = self.transpose_scores(
            self.value_layer(graph_elements))  # [B,N,S,H]

        attention_probs = tf.nn.softmax(query)
        context = tf.matmul(tf.expand_dims(attention_probs, 2), value)
        context = tf.reshape(context,
                             [batch_size, self.num_heads * self.units])

        if self.dropout > 0.:
            context = self.dropout_layer(context)

        return context

    def get_config(self):
        config = super(GlobalUpdate, self).get_config()
        config.update({"units": self.units, "num_heads": self.num_heads})
        return config



from typing import Callable, Dict, Hashable, Optional

import networkx as nx
import numpy as np
import rdkit.Chem

try:
    import tensorflow as tf
except ImportError:
    tf = None

from nfp.preprocessing import features
from nfp.preprocessing.preprocessor import Preprocessor
from nfp.preprocessing.tokenizer import Tokenizer


class MolPreprocessor(Preprocessor):
    def __init__(
        self,
        atom_features: Optional[Callable[[rdkit.Chem.Atom], Hashable]] = None,
        bond_features: Optional[Callable[[rdkit.Chem.Bond], Hashable]] = None,
        **kwargs,
    ) -> None:
        super(MolPreprocessor, self).__init__(**kwargs)

        self.atom_tokenizer = Tokenizer()
        self.bond_tokenizer = Tokenizer()

        if atom_features is None:
            atom_features = features.atom_features_v1

        if bond_features is None:
            bond_features = features.bond_features_v1

        self.atom_features = atom_features
        self.bond_features = bond_features

    def create_nx_graph(self, mol: rdkit.Chem.Mol, **kwargs) -> nx.DiGraph:
        g = nx.Graph(mol=mol)
        g.add_nodes_from(((atom.GetIdx(), {"atom": atom}) for atom in mol.GetAtoms()))
        g.add_edges_from(
            (
                (bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), {"bond": bond})
                for bond in mol.GetBonds()
            )
        )
        return nx.DiGraph(g)

    def get_edge_features(
        self, edge_data: list, max_num_edges
    ) -> Dict[str, np.ndarray]:
        bond_feature_matrix = np.zeros(max_num_edges, dtype=self.output_dtype)
        for n, (start_atom, end_atom, bond_dict) in enumerate(edge_data):
            flipped = start_atom == bond_dict["bond"].GetEndAtomIdx()
            bond_feature_matrix[n] = self.bond_tokenizer(
                self.bond_features(bond_dict["bond"], flipped=flipped)
            )

        return {"bond": bond_feature_matrix}

    def get_node_features(
        self, node_data: list, max_num_nodes
    ) -> Dict[str, np.ndarray]:
        atom_feature_matrix = np.zeros(max_num_nodes, dtype=self.output_dtype)
        for n, atom_dict in node_data:
            atom_feature_matrix[n] = self.atom_tokenizer(
                self.atom_features(atom_dict["atom"])
            )
        return {"atom": atom_feature_matrix}

    def get_graph_features(self, graph_data: dict) -> Dict[str, np.ndarray]:
        return {}

    @property
    def atom_classes(self) -> int:
        """The number of atom types found (includes the 0 null-atom type)"""
        return self.atom_tokenizer.num_classes + 1

    @property
    def bond_classes(self) -> int:
        """The number of bond types found (includes the 0 null-bond type)"""
        return self.bond_tokenizer.num_classes + 1

    @property
    def output_signature(self) -> Dict[str, tf.TensorSpec]:
        if tf is None:
            raise ImportError("Tensorflow was not found")
        return {
            "atom": tf.TensorSpec(shape=(None,), dtype=self.output_dtype),
            "bond": tf.TensorSpec(shape=(None,), dtype=self.output_dtype),
            "connectivity": tf.TensorSpec(shape=(None, 2), dtype=self.output_dtype),
        }

    @property
    def padding_values(self) -> Dict[str, tf.constant]:
        """Defaults to zero for each output"""
        if tf is None:
            raise ImportError("Tensorflow was not found")
        return {
            key: tf.constant(0, dtype=self.output_dtype)
            for key in self.output_signature.keys()
        }

    @property
    def tfrecord_features(self) -> Dict[str, tf.io.FixedLenFeature]:
        """For loading preprocessed inputs from a tf records file"""
        if tf is None:
            raise ImportError("Tensorflow was not found")
        return {
            key: tf.io.FixedLenFeature(
                [], dtype=self.output_dtype if len(val.shape) == 0 else tf.string
            )
            for key, val in self.output_signature.items()
        }


class SmilesPreprocessor(MolPreprocessor):
    def __init__(self, *args, explicit_hs: bool = True, **kwargs):
        super(SmilesPreprocessor, self).__init__(*args, **kwargs)
        self.explicit_hs = explicit_hs

    def create_nx_graph(self, smiles: str, *args, **kwargs) -> nx.DiGraph:
        mol = rdkit.Chem.MolFromSmiles(smiles)
        if self.explicit_hs:
            mol = rdkit.Chem.AddHs(mol)
        return super(SmilesPreprocessor, self).create_nx_graph(mol, *args, **kwargs)


class BondIndexPreprocessor(MolPreprocessor):
    def get_edge_features(
        self, edge_data: list, max_num_edges
    ) -> Dict[str, np.ndarray]:
        bond_indices = np.zeros(max_num_edges, dtype=self.output_dtype)
        for n, (_, _, edge_dict) in enumerate(edge_data):
            bond_indices[n] = edge_dict["bond"].GetIdx()
        edge_features = super(BondIndexPreprocessor, self).get_edge_features(
            edge_data, max_num_edges
        )
        return {"bond_indices": bond_indices, **edge_features}

    @property
    def output_signature(self) -> Dict[str, tf.TensorSpec]:
        if tf is None:
            raise ImportError("Tensorflow was not found")

        signature = super(BondIndexPreprocessor, self).output_signature
        signature["bond_indices"] = tf.TensorSpec(
            shape=(None,), dtype=self.output_dtype
        )
        return signature


class SmilesBondIndexPreprocessor(SmilesPreprocessor, BondIndexPreprocessor):
    pass



""" Keras loss functions will choke on NaN inputs, which we'll often have for
unspecified inputs. These are just a couple simple loss functions that mask NaN
values in both the test and predicted tensors when computing the cost. """

import tensorflow as tf


def masked_mean_squared_error(y_true, y_pred):
    mask = tf.math.is_finite(y_true)
    y_true_mask = tf.boolean_mask(tensor=y_true, mask=mask)
    y_pred_mask = tf.boolean_mask(tensor=y_pred, mask=mask)
    return tf.math.reduce_mean(tf.math.square(y_pred_mask - y_true_mask))


def masked_mean_absolute_error(y_true, y_pred):
    mask = tf.math.is_finite(y_true)
    y_true_mask = tf.boolean_mask(tensor=y_true, mask=mask)
    y_pred_mask = tf.boolean_mask(tensor=y_pred, mask=mask)
    return tf.math.reduce_mean(tf.math.abs(y_pred_mask - y_true_mask))


def masked_log_cosh(y_true, y_pred):
    mask = tf.math.is_finite(y_true)
    y_true_mask = tf.boolean_mask(tensor=y_true, mask=mask)
    y_pred_mask = tf.boolean_mask(tensor=y_pred, mask=mask)
    return tf.keras.losses.logcosh(y_pred_mask, y_true_mask)



from typing import Dict

import networkx as nx
import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None

from nfp.preprocessing.preprocessor import PreprocessorMultiGraph
from nfp.preprocessing.tokenizer import Tokenizer


class PymatgenPreprocessor(PreprocessorMultiGraph):
    def __init__(self, radius=None, num_neighbors=12, **kwargs):
        super(PymatgenPreprocessor, self).__init__(**kwargs)
        self.site_tokenizer = Tokenizer()
        self.radius = radius
        self.num_neighbors = num_neighbors

    def create_nx_graph(self, crystal, **kwargs) -> nx.MultiDiGraph:
        """ crystal should be a pymatgen.core.Structure object.
        """
        g = nx.MultiDiGraph(crystal=crystal)
        g.add_nodes_from(((i, {
            'site': site
        }) for i, site in enumerate(crystal.sites)))

        if self.radius is None:
            # Get the expected number of sites / volume, then find a radius
            # expected to yield 2x the desired number of neighbors
            desired_vol = (crystal.volume /
                           crystal.num_sites) * self.num_neighbors
            radius = 2 * (desired_vol / (4 * np.pi / 3)) ** (1 / 3)
        else:
            radius = self.radius

        for i, neighbors in enumerate(crystal.get_all_neighbors(radius)):
            if len(neighbors) < self.num_neighbors:
                raise RuntimeError(
                    f"Only {len(neighbors)} neighbors for site {i}")

            sorted_neighbors = sorted(neighbors,
                                      key=lambda x: x[1])[:self.num_neighbors]

            for _, distance, j, _ in sorted_neighbors:
                g.add_edge(i, j, distance=distance)

        return g

    def get_edge_features(self, edge_data: list,
                          max_num_edges) -> Dict[str, np.ndarray]:

        edge_feature_matrix = np.empty(max_num_edges, dtype='float32')
        edge_feature_matrix[:] = np.nan  # Initialize distances with nans

        for n, (_, _, edge_dict) in enumerate(edge_data):
            edge_feature_matrix[n] = edge_dict['distance']
        return {'distance': edge_feature_matrix}

    def get_node_features(self, node_data: list,
                          max_num_nodes) -> Dict[str, np.ndarray]:
        site_feature_matrix = np.zeros(max_num_nodes, dtype=self.output_dtype)
        for n, site_dict in node_data:
            site_feature_matrix[n] = self.site_tokenizer(
                self.site_features(site_dict['site']))
        return {'site': site_feature_matrix}

    def get_graph_features(self, graph_data: dict) -> Dict[str, np.ndarray]:
        return {}

    @property
    def site_classes(self):
        return self.site_tokenizer.num_classes + 1

    @staticmethod
    def site_features(site):
        species = site.as_dict()['species']
        assert len(species) == 1
        return species[0]['element']

    @property
    def output_signature(self) -> Dict[str, tf.TensorSpec]:
        if tf is None:
            raise ImportError('Tensorflow was not found')
        return {
            'site': tf.TensorSpec(shape=(None,), dtype=self.output_dtype),
            'distance': tf.TensorSpec(shape=(None,), dtype='float32'),
            'connectivity': tf.TensorSpec(shape=(None, 2),
                                          dtype=self.output_dtype)
        }

    @property
    def padding_values(self) -> Dict[str, tf.constant]:
        if tf is None:
            raise ImportError('Tensorflow was not found')
        return {
            'site': tf.constant(0, dtype=self.output_dtype),
            'distance': tf.constant(np.nan, dtype='float32'),
            'connectivity': tf.constant(0, dtype=self.output_dtype)
        }

    @property
    def tfrecord_features(self) -> Dict[str, tf.io.FixedLenFeature]:
        if tf is None:
            raise ImportError('Tensorflow was not found')
        return {
            key: tf.io.FixedLenFeature([], dtype=tf.string)
            for key in self.output_signature.keys()
        }



import json
from typing import Callable, Dict, Hashable, Optional, List

import networkx as nx
import numpy as np
import rdkit.Chem

try:
    import tensorflow as tf
except ImportError:
    tf = None

from nfp.preprocessing import features
from nfp.preprocessing.mol_preprocessor import MolPreprocessor, SmilesPreprocessor


class xTBPreprocessor(MolPreprocessor):
    def __init__(
        self,
        *args,
        explicit_hs: bool = True,
        xtb_atom_features: Optional[List[str]] = None,
        xtb_bond_features: Optional[List[str]] = None,
        xtb_mol_features: Optional[List[str]] = None,
        cutoff: float = 0.3,
        **kwargs
    ):
        super(xTBPreprocessor, self).__init__(*args, **kwargs)
        self.explicit_hs = explicit_hs
        self.cutoff = cutoff

        # update only bond features as we dont use rdkit
        self.bond_features = features.bond_features_wbo

        if xtb_atom_features is None:
            self.xtb_atom_features = [
                "mulliken charges",
                "cm5 charges",
                "FUKUI+",
                "FUKUI-",
                "FUKUIrad",
                "s proportion",
                "p proportion",
                "d proportion",
                "FOD",
                "FOD s proportion",
                "FOD p proportion",
                "FOD d proportion",
                "Dispersion coefficient C6",
                "Polarizability alpha",
            ]
        else:
            self.xtb_atom_features = xtb_atom_features

        if xtb_bond_features is None:
            self.xtb_bond_features = ["Wiberg matrix", "bond_dist"]
        else:
            self.xtb_bond_features = xtb_bond_features

        if xtb_mol_features is None:
            self.xtb_mol_features = [
                "total energy",
                "electronic energy",
                "HOMO",
                "LUMO",
            ]
        else:
            self.xtb_mol_features = xtb_mol_features

    def create_nx_graph(
        self, mol: rdkit.Chem.Mol, jsonfile: str, **kwargs
    ) -> nx.DiGraph:

        with open(jsonfile, "r") as f:
            json_data = json.load(f)

        # add hydrogens as wbo contains hydrogens and add xtb features to the graphs
        mol_data = {"mol_xtb": [json_data[prop] for prop in self.xtb_mol_features]}

        g = nx.Graph(mol=mol, **mol_data)
        for atom in mol.GetAtoms():
            atom_data = {
                "atom": atom,
                "atom_xtb": [
                    json_data[prop][atom.GetIdx()] for prop in self.xtb_atom_features
                ],
            }
            g.add_node(atom.GetIdx(), **atom_data)

        # add edges based on wilberg bond orders.
        wbo = np.array(json_data["Wiberg matrix"])
        edges_to_add = (
            (i, j)
            for i in range(len(wbo))
            for j in range(len(wbo))
            if wbo[i][j] > self.cutoff
        )

        for i, j in edges_to_add:

            edge_data = {
                "bondatoms": (mol.GetAtomWithIdx(i), mol.GetAtomWithIdx(j)),
                "bond_xtb": [json_data[prop][i][j] for prop in self.xtb_bond_features],
            }

            g.add_edge(i, j, **edge_data)

        return nx.DiGraph(g)

    def get_edge_features(
        self, edge_data: list, max_num_edges
    ) -> Dict[str, np.ndarray]:

        bond_feature_matrix = np.zeros(max_num_edges, dtype=self.output_dtype)
        bond_feature_matrix_xtb = np.zeros(
            (max_num_edges, len(self.xtb_bond_features)), dtype="float32"
        )

        for n, (start_atom, end_atom, bond_dict) in enumerate(edge_data):
            bond_feature_matrix[n] = self.bond_tokenizer(
                self.bond_features(start_atom, end_atom, bond_dict["bondatoms"])
            )
            bond_feature_matrix_xtb[n] = bond_dict["bond_xtb"]

        return {"bond": bond_feature_matrix, "bond_xtb": bond_feature_matrix_xtb}

    def get_node_features(
        self, node_data: list, max_num_nodes: int
    ) -> Dict[str, np.ndarray]:
        node_features = super().get_node_features(node_data, max_num_nodes)
        node_features["atom_xtb"] = np.zeros(
            [max_num_nodes, len(self.xtb_atom_features)], dtype="float32"
        )
        for n, atom_dict in node_data:
            node_features["atom_xtb"][n] = atom_dict["atom_xtb"]
        return node_features

    def get_graph_features(self, graph_data: dict) -> Dict[str, np.ndarray]:
        return {"mol_xtb": np.asarray(graph_data["mol_xtb"])}

    @property
    def output_signature(self) -> Dict[str, tf.TensorSpec]:
        output_signature = super().output_signature
        output_signature["atom_xtb"] = tf.TensorSpec(
            shape=(None, None), dtype="float32"
        )
        output_signature["bond_xtb"] = tf.TensorSpec(
            shape=(None, None), dtype="float32"
        )
        output_signature["mol_xtb"] = tf.TensorSpec(shape=(None,), dtype="float32")
        return output_signature

    @property
    def padding_values(self) -> Dict[str, tf.constant]:
        padding_values = super().padding_values
        padding_values["atom_xtb"] = tf.constant(np.nan, dtype="float32")
        padding_values["bond_xtb"] = tf.constant(np.nan, dtype="float32")
        padding_values["mol_xtb"] = tf.constant(np.nan, dtype="float32")
        return padding_values

    @property
    def tfrecord_features(self) -> Dict[str, tf.io.FixedLenFeature]:
        tfrecord_features = super().tfrecord_features
        tfrecord_features["atom_xtb"] = tf.io.FixedLenFeature(
            [],
            dtype="float32"
            if len(self.output_signature["atom_xtb"].shape) == 0
            else tf.string,
        )
        tfrecord_features["bond_xtb"] = tf.io.FixedLenFeature(
            [],
            dtype=self.output_dtype
            if len(self.output_signature["bond_xtb"].shape) == 0
            else tf.string,
        )
        tfrecord_features["mol_xtb"] = tf.io.FixedLenFeature(
            [],
            dtype=self.output_dtype
            if len(self.output_signature["mol_xtb"].shape) == 0
            else tf.string,
        )
        return tfrecord_features


class xTBSmilesPreprocessor(SmilesPreprocessor, xTBPreprocessor):
    pass



import json
import logging
import warnings
from abc import ABC, abstractmethod
from inspect import getmembers
from typing import Any, Dict, Optional

import networkx as nx
import numpy as np
import tensorflow as tf
from nfp.preprocessing.tokenizer import Tokenizer

logger = logging.getLogger(__name__)


class Preprocessor(ABC):
    """A base class for graph preprocessing

    Parameters
    ----------
    output_dtype
        A parameter used in child classes for determining the datatype of
        the returned arrays
    """

    def __init__(self, output_dtype: str = "int32"):
        self.output_dtype = output_dtype

    @abstractmethod
    def create_nx_graph(self, structure: Any, *args, **kwargs) -> nx.DiGraph:
        """Given an input structure object, convert it to a networkx digraph
        with node, edge, and graph features assigned.

        Parameters
        ----------
        structure
            Any input graph object
        kwargs
            keyword arguments passed from `__call__`, useful for specifying
            additional features in addition to the graph object.

        Returns
        -------
        nx.DiGraph
            A networkx graph with the node, edge, and graph features set
        """
        pass

    @abstractmethod
    def get_edge_features(
        self, edge_data: list, max_num_edges: int
    ) -> Dict[str, np.ndarray]:
        """Given a list of edge features from the nx.Graph, processes and
        concatenates them to an array.

        Parameters
        ----------
        edge_data
            A list of edge data generated by `nx_graph.edges(data=True)`
        max_num_edges
            If desired, this function should pad to a maximum number of edges
            passed from the `__call__` function.

        Returns
        -------
        Dict[str, np.ndarray]
            a dictionary of feature, array pairs, where array contains features for
            all edges in the graph.
        """
        pass

    @abstractmethod
    def get_node_features(
        self, node_data: list, max_num_nodes: int
    ) -> Dict[str, np.ndarray]:
        """Given a list of node features from the nx.Graph, processes and
        concatenates them to an array.

        Parameters
        ----------
        node_data
            A list of edge data generated by `nx_graph.nodes(data=True)`
        max_num_nodes
            If desired, this function should pad to a maximum number of nodes
            passed from the `__call__` function.

        Returns
        -------
        Dict[str, np.ndarray]
            a dictionary of feature, array pairs, where array contains features for
            all nodes in the graph.
        """
        pass

    @abstractmethod
    def get_graph_features(self, graph_data: dict) -> Dict[str, np.ndarray]:
        """Process the nx.graph features into a dictionary of arrays.

        Parameters
        ----------
        graph_data
            A dictionary of graph data generated by `nx_graph.graph`

        Returns
        -------
        Dict[str, np.ndarray]
            a dictionary of features for the graph
        """
        pass

    @staticmethod
    def get_connectivity(
        graph: nx.DiGraph, max_num_edges: int
    ) -> Dict[str, np.ndarray]:
        """Get the graph connectivity from the networkx graph

        Parameters
        ----------
        graph
            The input graph
        max_num_edges
            len(graph.edges), or the specified maximum number of graph edges

        Returns
        -------
        Dict[str, np.ndarray]
            A dictionary of with the single 'connectivity' key, containing an (n,2)
            array of (node_index, node_index) pairs indicating the start and end
            nodes for each edge.
        """
        connectivity = np.zeros((max_num_edges, 2), dtype="int64")
        if len(graph.edges) > 0:  # Handle odd case with no edges
            connectivity[: len(graph.edges)] = np.asarray(graph.edges)
        return {"connectivity": connectivity}

    def __call__(
        self,
        structure: Any,
        *args,
        train: bool = False,
        max_num_nodes: Optional[int] = None,
        max_num_edges: Optional[int] = None,
        **kwargs,
    ) -> Dict[str, np.ndarray]:
        """Convert an input graph structure into a featurized set of node, edge,
         and graph-level features.

        Parameters
        ----------
        structure
            An input graph structure (i.e., molecule, crystal, etc.)
        train
            A training flag passed to `Tokenizer` member attributes
        max_num_nodes
            A size attribute passed to `get_node_features`, defaults to the
            number of nodes in the current graph
        max_num_edges
            A size attribute passed to `get_edge_features`, defaults to the
            number of edges in the current graph
        kwargs
            Additional features or parameters passed to `construct_nx_graph`

        Returns
        -------
        Dict[str, np.ndarray]
            A dictionary of key, array pairs as a single sample.
        """
        nx_graph = self.create_nx_graph(structure, *args, **kwargs)

        max_num_edges = len(nx_graph.edges) if max_num_edges is None else max_num_edges
        assert (
            len(nx_graph.edges) <= max_num_edges
        ), "max_num_edges too small for given input"

        max_num_nodes = len(nx_graph.nodes) if max_num_nodes is None else max_num_nodes
        assert (
            len(nx_graph.nodes) <= max_num_nodes
        ), "max_num_nodes too small for given input"

        # Make sure that Tokenizer classes are correctly initialized
        for _, tokenizer in getmembers(self, lambda x: type(x) == Tokenizer):
            tokenizer.train = train

        node_features = self.get_node_features(nx_graph.nodes(data=True), max_num_nodes)
        edge_features = self.get_edge_features(nx_graph.edges(data=True), max_num_edges)
        graph_features = self.get_graph_features(nx_graph.graph)
        connectivity = self.get_connectivity(nx_graph, max_num_edges)

        return {**node_features, **edge_features, **graph_features, **connectivity}

    def construct_feature_matrices(
        self, *args, train=False, **kwargs
    ) -> Dict[str, np.ndarray]:
        """
        .. deprecated:: 0.3.0
            `construct_feature_matrices` will be removed in 0.4.0, use
            `__call__` instead
        """
        warnings.warn(
            "construct_feature_matrices is deprecated, use `call` instead as "
            "of nfp 0.4.0",
            DeprecationWarning,
        )
        return self(*args, train=train, **kwargs)

    def to_json(self, filename: str) -> None:
        """Serialize the classes's data to a json file"""
        with open(filename, "w") as f:
            json.dump(self, f, default=lambda x: x.__dict__)

    def from_json(self, filename: str) -> None:
        """Set's the class's data with attributes taken from the save file"""
        with open(filename, "r") as f:
            json_data = json.load(f)
        load_from_json(self, json_data)

    @property
    @abstractmethod
    def output_signature(self) -> Dict[str, tf.TensorSpec]:
        pass

    @property
    @abstractmethod
    def padding_values(self) -> Dict[str, tf.constant]:
        pass

    @property
    @abstractmethod
    def tfrecord_features(self) -> Dict[str, tf.io.FixedLenFeature]:
        pass


class PreprocessorMultiGraph(Preprocessor, ABC):
    """Class to handle graphs with parallel edges and self-loops"""

    @abstractmethod
    def create_nx_graph(self, structure: Any, **kwargs) -> nx.MultiDiGraph:
        pass

    @staticmethod
    def get_connectivity(
        graph: nx.DiGraph, max_num_edges: int
    ) -> Dict[str, np.ndarray]:
        # Don't include keys in the connectivity matrix
        connectivity = np.zeros((max_num_edges, 2), dtype="int64")
        if len(graph.edges) > 0:  # Handle odd case with no edges
            connectivity[: len(graph.edges)] = np.asarray(graph.edges)[:, :2]
        return {"connectivity": connectivity}


def load_from_json(obj: Any, data: Dict):
    """Function to set member attributes from json data recursively.

    Parameters
    ----------
    obj
        the class to initialize
    data
        a dictionary of potentially nested attribute - value pairs

    Returns
    -------
    Any
        The object, with attributes set to those from the data file.
    """

    for key, val in obj.__dict__.items():
        try:
            if isinstance(val, type(data[key])):
                obj.__dict__[key] = data[key]
            elif hasattr(val, "__dict__"):
                load_from_json(val, data[key])

        except KeyError:
            logger.warning(
                f"{key} not found in JSON file, it may have been created with"
                " an older nfp version"
            )



import tensorflow as tf
from tensorflow.keras import layers as tf_layers


class RBFExpansion(tf_layers.Layer):
    def __init__(self,
                 dimension=128,
                 init_gap=10,
                 init_max_distance=7,
                 trainable=False):
        """ Layer to calculate radial basis function 'embeddings' for a continuous input variable. The width and
        location of each bin can be optionally trained. Essentially equivalent to a 1-hot embedding for a continuous
        variable.

        Parameters
        ----------
        dimension: The total number of distance bins
        init_gap: The initial width of each gaussian distribution
        init_max_distance: the initial maximum value of the continuous variable
        trainable: Whether the centers and gap parameters should be added as trainable NN parameters.
        """
        super(RBFExpansion, self).__init__()
        self.init_gap = init_gap
        self.init_max_distance = init_max_distance
        self.dimension = dimension
        self.trainable = trainable

    def build(self, input_shape):
        self.centers = tf.Variable(
            name='centers',
            initial_value=tf.range(
                0, self.init_max_distance,
                delta=self.init_max_distance / self.dimension),
            trainable=self.trainable,
            dtype=tf.float32)

        self.gap = tf.Variable(name='gap',
                               initial_value=tf.constant(self.init_gap,
                                                         dtype=tf.float32),
                               trainable=self.trainable,
                               dtype=tf.float32)

    def call(self, inputs, **kwargs):
        distances = tf.where(tf.math.is_nan(inputs),
                             tf.zeros_like(inputs, dtype=inputs.dtype), inputs)
        offset = tf.expand_dims(distances, -1) - tf.cast(
            self.centers, inputs.dtype)
        logits = -self.gap * offset ** 2
        return tf.exp(logits)

    def compute_mask(self, inputs, mask=None):
        return tf.logical_not(tf.math.is_nan(inputs))

    def get_config(self):
        return {
            'init_gap': self.init_gap,
            'init_max_distance': self.init_max_distance,
            'dimension': self.dimension,
            'trainable': self.trainable
        }


def batched_segment_op(data,
                       segment_ids,
                       num_segments,
                       data_mask=None,
                       reduction='sum'):
    """Flattens data and segment_ids containing a batch dimension for
    tf.math.segment* operations. Includes support for masking.

    Arguments:
        data: tensor of shape [B, L, F], where B is the batch size, L is the
            length, and F is a feature dimension
        segment_ids: tensor of shape [B, L] containing up to N segments
        num_segments: N, integer
        data_mask: boolean tensor of shape [B, L] masking the input data
        reduction: string for specific tf.math.unsorted_segment_* function

    """

    if data_mask is None:
        data_mask = tf.ones(tf.shape(data)[:-1], dtype=tf.bool)

    # Prior to flattening, offset rows of segment_ids to preserve batches
    batch_size = tf.shape(data, out_type=segment_ids.dtype)[0]
    offsets = tf.range(batch_size, dtype=segment_ids.dtype) * num_segments
    ids_offset = segment_ids + tf.expand_dims(offsets, 1)

    # Mask and flatten the data and segment_ids
    flat_data = tf.boolean_mask(data, data_mask)
    flat_ids = tf.boolean_mask(ids_offset, data_mask)

    reduction = getattr(tf.math, f'unsorted_segment_{reduction}')

    # Perform the segment operation on the flattened vectors, and reshape the
    # result
    reduced_data = reduction(flat_data, flat_ids, num_segments * batch_size)
    return tf.reshape(reduced_data, [batch_size, num_segments, data.shape[-1]])


# class Slice(tf_layers.Layer):
#     def __init__(self, slice_obj, *args, **kwargs):
#         super(Slice, self).__init__(*args, **kwargs)
#         self.slice_obj = slice_obj
#         self.supports_masking = True
#
#     def call(self, inputs, mask=None):
#         return inputs[self.slice_obj]
#
#     def get_config(self):
#         return {'slice_obj': str(self.slice_obj)}
#
#     @classmethod
#     def from_config(cls, config):
#         config['slice_obj'] = eval(config['slice_obj'])
#         return cls(**config)


class Gather(tf_layers.Layer):
    def call(self, inputs, mask=None, **kwargs):
        reference, indices = inputs
        return tf.gather(reference, indices, batch_dims=1)


class Reduce(tf_layers.Layer):
    def __init__(self, reduction='sum', *args, **kwargs):
        super(Reduce, self).__init__(*args, **kwargs)
        self.reduction = reduction

    def compute_output_shape(self, input_shape):
        data_shape, _, target_shape = input_shape
        return [data_shape[0], target_shape[1], data_shape[-1]]

    def call(self, inputs, mask=None, **kwargs):
        data, segment_ids, target, data_mask = self._parse_inputs_and_mask(
            inputs, mask)
        num_segments = tf.shape(target, out_type=segment_ids.dtype)[1]
        return batched_segment_op(data,
                                  segment_ids,
                                  num_segments,
                                  data_mask=data_mask,
                                  reduction=self.reduction)

    def get_config(self):
        return {'reduction': self.reduction}

    @staticmethod
    def _parse_inputs_and_mask(inputs, mask=None):
        data, segment_ids, target = inputs

        # Handle missing masks
        if mask is not None:
            data_mask = mask[0]
        else:
            data_mask = None

        return data, segment_ids, target, data_mask


class ConcatDense(tf_layers.Layer):
    """ Layer to combine the concatenation and two dense layers. Just useful as a common operation in the graph
    layers """

    def __init__(self, **kwargs):
        super(ConcatDense, self).__init__(**kwargs)
        self.supports_masking = True

    def build(self, input_shape):
        num_features = input_shape[0][-1]
        self.concat = tf_layers.Concatenate()
        self.dense1 = tf_layers.Dense(2 * num_features, activation='relu')
        self.dense2 = tf_layers.Dense(num_features)

    def call(self, inputs, mask=None, **kwargs):
        output = self.concat(inputs)
        output = self.dense1(output)
        output = self.dense2(output)
        return output

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None
        else:
            return tf.math.reduce_all(tf.stack(mask), axis=0)


class Tile(tf_layers.Layer):
    def __init__(self, **kwargs):
        super(Tile, self).__init__(**kwargs)
        self.supports_masking = True

    def call(self, inputs, mask=None, **kwargs):
        global_state, target = inputs
        target_shape = tf.shape(target)[1]  # number of edges or nodes
        expanded = tf.expand_dims(global_state, 1)
        return tf.tile(expanded, tf.stack([1, target_shape, 1]))

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None
        else:
            return mask[1]



class Tokenizer(object):
    """A class to turn arbitrary inputs into integer classes."""

    def __init__(self):
        # the default class for an unseen entry during test-time
        self._data = {"unk": 1}
        self.num_classes = 1
        self.train = True
        self.unknown = []

    def __call__(self, item):
        """Check to see if the Tokenizer has seen `item` before, and if so,
        return the integer class associated with it. Otherwise, if we're
        training, create a new integer class, otherwise return the 'unknown'
        class.

        """
        try:
            return self._data[item]

        except KeyError:
            if self.train:
                self._add_token(item)
                return self(item)

            else:
                # Record the unknown item, then return the unknown label
                self.unknown += [item]
                return self._data["unk"]

    def _add_token(self, item):
        self.num_classes += 1
        self._data[item] = self.num_classes



# The rest of the methods in this module are specific functions for computing
# atom and bond features. New ones can be easily added though, and these are
# passed directly to the Preprocessor class.


def get_ring_size(obj, max_size=12):
    if not obj.IsInRing():
        return 0
    else:
        for i in range(max_size):
            if obj.IsInRingSize(i):
                return i
        else:
            return "max"


def atom_features_v1(atom):
    """Return an integer hash representing the atom type"""

    return str(
        (
            atom.GetSymbol(),
            atom.GetDegree(),
            atom.GetTotalNumHs(),
            atom.GetImplicitValence(),
            atom.GetIsAromatic(),
        )
    )


def atom_features_v2(atom):
    props = [
        "GetChiralTag",
        "GetDegree",
        "GetExplicitValence",
        "GetFormalCharge",
        "GetHybridization",
        "GetImplicitValence",
        "GetIsAromatic",
        "GetNoImplicit",
        "GetNumExplicitHs",
        "GetNumImplicitHs",
        "GetNumRadicalElectrons",
        "GetSymbol",
        "GetTotalDegree",
        "GetTotalNumHs",
        "GetTotalValence",
    ]

    atom_type = [getattr(atom, prop)() for prop in props]
    atom_type += [get_ring_size(atom)]

    return str(tuple(atom_type))


def bond_features_v1(bond, **kwargs):
    """Return an integer hash representing the bond type.

    flipped : bool
        Only valid for 'v3' version, whether to swap the begin and end atom
        types

    """

    return str(
        (
            bond.GetBondType(),
            bond.GetIsConjugated(),
            bond.IsInRing(),
            sorted([bond.GetBeginAtom().GetSymbol(), bond.GetEndAtom().GetSymbol()]),
        )
    )


def bond_features_v2(bond, **kwargs):
    return str(
        (
            bond.GetBondType(),
            bond.GetIsConjugated(),
            bond.GetStereo(),
            get_ring_size(bond),
            sorted([bond.GetBeginAtom().GetSymbol(), bond.GetEndAtom().GetSymbol()]),
        )
    )


def bond_features_v3(bond, flipped=False):
    if not flipped:
        start_atom = atom_features_v1(bond.GetBeginAtom())
        end_atom = atom_features_v1(bond.GetEndAtom())

    else:
        start_atom = atom_features_v1(bond.GetEndAtom())
        end_atom = atom_features_v1(bond.GetBeginAtom())

    return str(
        (
            bond.GetBondType(),
            bond.GetIsConjugated(),
            bond.GetStereo(),
            get_ring_size(bond),
            bond.GetEndAtom().GetSymbol(),
            start_atom,
            end_atom,
        )
    )


def bond_features_wbo(start_atom, end_atom, bondatoms):

    start_atom_symbol = bondatoms[0].GetSymbol()
    end_atom_symbol = bondatoms[1].GetSymbol()

    return str((start_atom_symbol, end_atom_symbol))



from .layers import *
from .models import *
from .preprocessing import *


class MissingDependencyException(RuntimeError):
    pass


custom_objects = {
    'Gather': Gather,
    'Reduce': Reduce,
    'masked_mean_squared_error': masked_mean_squared_error,
    'masked_mean_absolute_error': masked_mean_absolute_error,
    'masked_log_cosh': masked_log_cosh,
    'EdgeUpdate': EdgeUpdate,
    'NodeUpdate': NodeUpdate,
    'GlobalUpdate': GlobalUpdate
}

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions




# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-01-25T13:20:06-0700",
 "dirty": false,
 "error": null,
 "full-revisionid": "e19d47640420a1c5a6ae7d708af97ba2b13178e7",
 "version": "0.3.8"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
